"""GnuCash utility functions.

Mostly wrappers around piecash functions.
"""
from contextlib import contextmanager

from werkzeug.exceptions import NotFound, Locked
from flask import request
import piecash
import sqlalchemy


class AccessDenied(Exception):
    """Database access was denied."""

    pass


class AccountNotFound(NotFound):
    """GnuCash Account was not found."""

    def __init__(self, fullname, *args, **kwargs):
        """Create error.

        :param fullname: Full name of the non-existent account
        :returns: New Exception

        """
        super().__init__(*args, **kwargs)
        self.account_name = fullname


class DatabaseLocked(Locked):
    """The Database is locked."""

    def __init__(self, *args, **kwargs):
        """Create error.

        :returns: New Exception

        """
        super().__init__(*args, **kwargs)


@contextmanager
def open_book(*args, **kwargs):
    """Open GnuCash book in the configured database.

    Should be used as context manager.

    Wraps around `piecash.open_book`. Parameters are passed to that function.

    :param open_if_lock: If not provided explicitly, this is read from `request.args`
    :returns: The book
    :raises DatabaseLocked: If the databased is being accessed by someone else and
      `open_if_lock` is not `True`
    :raises AccessDenied: If access to the database is denied by the SQL server.

    """
    try:
        if "open_if_lock" not in kwargs:
            kwargs["open_if_lock"] = request.args.get(
                "open_if_lock", default=False, type=bool
            )

        with piecash.open_book(*args, **kwargs) as book:
            yield book

    except piecash.GnucashException as e:
        if "Lock on the file" in str(e):
            raise DatabaseLocked()
        else:
            raise e

    except sqlalchemy.exc.OperationalError as e:
        if "Access denied" in str(e):
            raise AccessDenied from e
        else:
            raise e


def get_account(book, *args, **kwargs):
    """Get account in the book based on given filters.

    Wraps around `piecash.core.book.Book.accounts.all`. All parameters are passed to
    that function.

    :param book: The book containing the account.
    :returns:

    """
    try:
        return book.accounts.get(*args, **kwargs)
    except KeyError:
        raise AccountNotFound(*args, **kwargs)




from werkzeug.datastructures import CallbackDict
from flask.sessions import SessionInterface, SessionMixin

from Crypto.Cipher import AES
import json
from json import JSONEncoder, JSONDecoder
import base64
import zlib

class EncryptedSession(CallbackDict, SessionMixin):

    def __init__(self, initial=None):
        def on_update(self):
            self.modified = True
        CallbackDict.__init__(self, initial, on_update)
        self.modified = False


class EncryptedSessionInterface(SessionInterface):
    session_class = EncryptedSession
    compress_threshold = 1024

    def open_session(self, app, request):
        '''
        @param py: Flask py
        @param request: Flask HTTP Request
        @summary: Sets the current session from the request's session cooke. This overrides the default
        Flask implementation, adding AES decryption of the client-side session cookie.
        '''

        # Get the session cookie
        session_cookie = request.cookies.get(app.session_cookie_name)
        if not session_cookie:
            return self.session_class()

        # Get the crypto key
        crypto_key = app.config['SESSION_CRYPTO_KEY'] if 'SESSION_CRYPTO_KEY' in app.config else app.crypto_key

        # Split the session cookie : <z|u>.<base64 cipher text>.<base64 mac>.<base64 nonce>
        itup = session_cookie.split(".")
        if (len (itup) is not 4):
            return self.session_class()     # Session cookie not in the right format

        try:

            # Compressed data?
            if (itup[0] == 'z'):            # session cookie for compressed data starts with "z."
                is_compressed = True
            else:
                is_compressed = False

            # Decode the cookie parts from base64
            ciphertext = base64.b64decode (bytes(itup[1], 'utf-8'))
            mac = base64.b64decode (bytes(itup[2], 'utf-8'))
            nonce = base64.b64decode (bytes(itup[3], 'utf-8'))

            # Decrypt
            cipher = AES.new (crypto_key, AES.MODE_EAX, nonce)
            data = cipher.decrypt_and_verify (ciphertext, mac)

            # Convert back to a dict and pass that onto the session
            if is_compressed:
                data = zlib.decompress (data)
            session_dict = json.loads (str (data, 'utf-8'), cls=BinaryAwareJSONDecoder)

            return self.session_class (session_dict)

        except ValueError:
            return self.session_class()

    def save_session(self, app, session, response):
        '''
        @param py: Flask py
        @param session: Flask / Werkzeug Session
        @param response: Flask HTTP Response
        @summary: Saves the current session. This overrides the default Flask implementation, adding
        AES encryption of the client-side session cookie.
        '''

        domain = self.get_cookie_domain(app)
        if not session:
            if session.modified:
                response.delete_cookie (app.session_cookie_name, domain=domain)
            return
        expires = self.get_expiration_time(app, session)

        # Decide whether to compress
        bdict = bytes (json.dumps (dict (session), cls=BinaryAwareJSONEncoder), 'utf-8')
        if (len (bdict) > self.compress_threshold):
            prefix = "z"                                    # session cookie for compressed data starts with "z."
            bdict = zlib.compress (bdict)
        else:
            prefix = "u"                                    # session cookie for uncompressed data starts with "u."

        # Get the crypto key
        crypto_key = app.config['SESSION_CRYPTO_KEY'] if 'SESSION_CRYPTO_KEY' in app.config else app.crypto_key

        # Encrypt using AES in EAX mode
        cipher = AES.new (crypto_key, AES.MODE_EAX)
        ciphertext, mac = cipher.encrypt_and_digest (bdict)
        nonce = cipher.nonce

        # Convert the ciphertext, mac, and nonce to base64
        b64_ciphertext = base64.b64encode (ciphertext)
        b64_mac = base64.b64encode (mac)
        b64_nonce = base64.b64encode (nonce)

        # Create the session cookie as <u|z>.<base64 cipher text>.<base64 mac>.<base64 nonce>
        tup = [prefix, b64_ciphertext.decode(), b64_mac.decode(), b64_nonce.decode()]
        session_cookie = ".".join(tup)

        # Set the session cookie
        response.set_cookie(app.session_cookie_name, session_cookie,
                            expires=expires, httponly=True,
                            domain=domain)


class BinaryAwareJSONEncoder(JSONEncoder):
    """ 
    Converts a python object, where binary data is converted into an object
    that can be decoded using the BinaryAwareJSONDecoder.
    """
    def default(self, obj):
        if isinstance(obj, bytes):
            return {
                '__type__' : 'bytes',
                'b' : base64.b64encode (obj).decode ()
            }   

        else:
            return JSONEncoder.default(self, obj)

class BinaryAwareJSONDecoder(JSONDecoder):
    """ 
    Converts a json string, where binary data was converted into objects form
    using the BinaryAwareJSONEncoder, back into a python object.
    """

    def __init__(self):
            JSONDecoder.__init__(self, object_hook=self.dict_to_object)

    def dict_to_object(self, d): 
        if '__type__' not in d:
            return d

        typ = d.pop('__type__')
        if typ == 'bytes':
            return base64.b64decode (bytes (d['b'], 'utf-8'))
        else:
            # Oops... better put this back together.
            d['__type__'] = typ
            return d



"""Flask app for GnuCash Web."""
import os
from importlib import metadata

from flask import Flask, redirect, url_for
from flask.cli import FlaskGroup
import click

from . import auth, book, commodities
from .utils import jinja as jinja_utils
from .config import GnuCashWebConfig

from encrypted_session import EncryptedSessionInterface


def create_app(test_config=None):
    """Create Flask app.

    Flask App factory.

    :param test_config: Optional testing config file
    :returns:

    """
    app = Flask('gnucash_web')
    app.config = GnuCashWebConfig(app)

    if not app.debug:
        app.logger.setLevel(app.config.LOG_LEVEL)

    # We encrypt the session-cookie, so the DB-password is not stored in plaintext
    # when using AUTH_MECHANISM == 'passthrough'.
    app.session_interface = EncryptedSessionInterface()

    if test_config is None:
        # load the instance config, if it exists, when not testing
        app.config.from_pyfile('config.py', silent=True)
    else:
        # load the test config if passed in
        app.config.from_mapping(test_config)

    # ensure the instance folder exists
    try:
        os.makedirs(app.instance_path)
    except OSError:
        pass

    app.jinja_env.autoescape = True
    app.jinja_env.filters['display'] = jinja_utils.safe_display_string
    app.jinja_env.filters['cssescape'] = jinja_utils.css_escape
    app.jinja_env.filters['parentaccounts'] = jinja_utils.parent_accounts
    app.jinja_env.filters['money'] = jinja_utils.money
    app.jinja_env.filters['accounturl'] = jinja_utils.account_url
    app.jinja_env.filters['full_account_names'] = jinja_utils.full_account_names
    app.jinja_env.filters['contrasplits'] = jinja_utils.contra_splits
    app.jinja_env.filters['nth'] = jinja_utils.nth
    app.jinja_env.globals['is_authenticated'] = auth.is_authenticated

    with open('version.txt') as version:
        app.jinja_env.globals['pkg_version'] = version.read().strip()

    app.register_blueprint(auth.bp)
    app.register_blueprint(book.bp)
    app.register_blueprint(commodities.bp)

    @app.route('/')
    def index():
        return redirect(url_for('book.show_account'))

    return app


@click.group(cls=FlaskGroup, create_app=create_app)
@click.option('--username')
@click.option('--password')
def cli(username, password):
    """Perform maintenance tasks on local GnuCash Web installation."""



"""Manage commodities in the GnuCash database."""
from flask import (
    Blueprint,
    current_app as app,
)
import click
from babel import numbers
from piecash.core.commodity import Price

from .utils.gnucash import open_book

bp = Blueprint("commodities", __name__, url_prefix="/commodities")


def latest_price(commodity):
    """Get latest known price for commodity.

    :param commodity: The commodity in question. Must be different from the books
      default currency

    :returns: Latest known price of the commodity in the books default currency

    """
    return commodity.prices.order_by(Price.date.desc()).limit(1).first()


def format_price(price):
    """Format price with currency according to current locale.

    :param price: The price in question
    :returns: Human-readable string

    """
    return numbers.format_currency(price.value, price.currency.mnemonic)


@bp.cli.command("list")
@click.option("--namespace", help="Filter by namespace")
@click.pass_context
def list_commodities(ctx, namespace):
    """List all used commodities.

    :param ctx: Click application context
    :param namespace: Filter by namespace

    """
    opts = ctx.find_root().params

    with open_book(
        uri_conn=app.config.DB_URI(opts.get("username"), opts.get("password")),
        readonly=True,
        open_if_lock=True,
    ) as book:
        if namespace:
            commodities = book.commodities(namespace=namespace)
        else:
            commodities = book.commodities

        for commodity in commodities:
            print(f"{commodity.namespace}:{commodity.mnemonic} ({commodity.cusip})")
            print(f"  Traded fraction: {1/commodity.fraction}")

            if commodity.quote_flag:
                print(f"  Quote Source: {commodity.quote_source} (ignored)")

            price = latest_price(commodity)
            if price:
                print(
                    f"  Latest known price: {format_price(price)}"
                    f" ({price.source}@{price.date})"
                )


@bp.cli.command("update_prices")
@click.pass_context
def update_prices(ctx):
    """Update prices for all commodities for which it is enabled.

    :param ctx: Click application context

    """
    opts = ctx.find_root().params

    with open_book(
        uri_conn=app.config.DB_URI(opts.get("username"), opts.get("password")),
        readonly=False,
        do_backup=False,
        open_if_lock=False,
    ) as book:

        old_prices = {}

        # Update prices (relative to book.default_currency) of all relevant
        # commodities
        for commodity in book.commodities:
            old_prices[commodity] = latest_price(commodity)
            if commodity.quote_flag and commodity != book.default_currency:
                commodity.update_prices()

        # Fetched prices are only visible after save, so we do it now and print the
        # changes later
        book.save()

        # Print price changes
        for commodity in book.commodities:
            old_price = old_prices[commodity]
            new_price = latest_price(commodity)
            if new_price:
                if old_price and new_price.date > old_price.date:
                    print(
                        f"New price for {commodity.mnemonic}:"
                        f" {format_price(old_price)}@{old_price.date}"
                        f" -> {format_price(new_price)}@{new_price.date}"
                    )
                else:
                    print(
                        f"Price for {commodity.mnemonic}:"
                        f" {format_price(new_price)}@{new_price.date}"
                    )



from functools import wraps

from flask import Blueprint, render_template, url_for, request, redirect, session
from flask import current_app as app
from sqlalchemy.exc import OperationalError

from .utils.gnucash import open_book, AccessDenied

bp = Blueprint('auth', __name__, url_prefix='/auth')

@bp.app_errorhandler(AccessDenied)
def handle_account_not_found(e: AccessDenied):
    end_session()
    return redirect(url_for('auth.login'))


def get_db_credentials():
    if not app.config.AUTH_MECHANISM:
        return (None, None)
    elif app.config.AUTH_MECHANISM == 'passthrough':
        return (session['username'], session['password'])
    else:
        raise NotImplementedError('Only passthrough auth is currently supported')

def authenticate(username, password):
    if not app.config.AUTH_MECHANISM:
        return True
    elif app.config.AUTH_MECHANISM == 'passthrough':
        try:
            # Check authn by attempting to connect to database
            with open_book(uri_conn=app.config.DB_URI(username, password),
                           readonly=True, open_if_lock=True) as book:
                app.logger.debug(f'Authentication succeeded for {username}')
                session['username'] = username
                session['password'] = password
                return True
        except AccessDenied:
            app.logger.debug(f'Authentication failed for {username}')
            return False
    else:
        raise NotImplementedError('Only passthrough auth is currently supported')

def end_session():
    if not app.config.AUTH_MECHANISM:
        pass
    elif app.config.AUTH_MECHANISM == 'passthrough':
        app.logger.debug(f'Logged out {session["username"]}')
        del session['username']
        del session['password']
    else:
        raise NotImplementedError('Only passthrough auth is currently supported')

def is_authenticated():
    return not app.config.AUTH_MECHANISM or 'username' in session


# Decorator for view functions
def requires_auth(func):
    @wraps(func)
    def inner(*args, **kwargs):
        if is_authenticated():
            return func(*args, **kwargs)
        else:
            return redirect(url_for('auth.login', return_url=request.url))

    return inner


@bp.route('/login', methods=['GET', 'POST'])
def login():
    if request.method == 'POST':
       authenticate(request.form['username'], request.form['password'])
       session.permanent = True
       return redirect(request.args.get('return_url') or url_for('.login'))
    else:
        if is_authenticated():
            return render_template('user.j2', username=session.get('username', 'no one'))
        else:
            return render_template('login.j2')

@bp.route('/logout', methods=['POST'])
def logout():
    end_session()
    return redirect(url_for('auth.login'))



"""Default GnuCash Web configuration."""
import logging
import os

SECRET_KEY = bytes.fromhex(os.getenv('SECRET_KEY', '00000000'))

LOG_LEVEL = logging.getLevelName(os.getenv('LOG_LEVEL', 'WARN'))

DB_DRIVER = os.getenv('DB_DRIVER', 'sqlite')
DB_NAME = os.getenv('DB_NAME', 'db/gnucash.sqlite')
DB_HOST = os.getenv('DB_HOST', 'localhost')

AUTH_MECHANISM = os.getenv('AUTH_MECHANISM')

TRANSACTION_PAGE_LENGTH = int(os.getenv('TRANSACTION_PAGE_LENGTH', 25))
PRESELECTED_CONTRA_ACCOUNT = os.getenv('PRESELECTED_CONTRA_ACCOUNT')
